/*****************************************************************************
//
//  Copyright (c) 2000-2003  Broadcom Corporation
//  All Rights Reserved
//  No portions of this material may be reproduced in any form without the
//  written permission of:
//          Broadcom Corporation
//          16215 Alton Parkway
//          Irvine, California 92619
//  All information contained in this document is Broadcom Corporation
//  company private, proprietary, and trade secret.
//
******************************************************************************
//
//  Filename:       fwsyscall.c
//  Author:         seanl
//  Creation Date:  10/10/03
//                  The system call for kill process and firmware flashing 
//                  functions (from syscall.c and dltftp.c)
//
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <pwd.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <dirent.h>
#include <net/if.h>
#include <ctype.h>

#include "fwsyscall.h"
#define  BCMTAG_EXE_USE
#include "bcmTag.h"

extern char **environ;
extern char glbIfName[];

#define READ_BUF_SIZE       128
#define CMD_LEN             256


/***************************************************************************
// Function Name: bcmSocketIfPid().
// Description  : return the socket if pid
// Parameters   : none.
// Returns      : pid
****************************************************************************/
int bcmSocketIfPid(void)
{
   int pid = 0;
   char buffer[CMD_LEN];
   char *ptr, *ptr2;
   FILE *fs;


   if (glbIfName[0] == '\0')
      return pid;    

   bcmSystem("ps > /var/pslist");
   fs = fopen("/var/pslist", "r");

   if (strstr(glbIfName, "nas") != NULL) {    // its a MER 
      while (fgets(buffer, CMD_LEN, fs) > 0) 
         if ((ptr = strstr(buffer, "dhcpc")) != NULL && (ptr2 = strstr(buffer, glbIfName)) != NULL) { // found command line 
            *(ptr2+5) = '\0';     // 'nas??\0', null the cmd
            pid = bcmGetPid(ptr);
         }
   }
   else if (strstr(glbIfName, "ppp") != NULL) { // its pppoe or pppoa
      char endCmd[5];
      endCmd[0] = ' ';           // eg. ' 35'
      strcpy(&endCmd[1], &glbIfName[3]);
      while (fgets(buffer, CMD_LEN, fs) > 0) 
         if ((ptr = strstr(buffer, "pppd")) != NULL && (ptr2 = strstr(buffer, endCmd)) != NULL) {// found command lin
            *(ptr2+3) = '\0';     // ' 35\0' null the cmd
            pid =bcmGetPid(ptr);
         }
   }
   
   fclose(fs);
   bcmSystem("rm /var/pslist");

   if (strstr(glbIfName, "ppp") != NULL)     // only do it if it is ppp
   {
      // add the default route with the ifc the socket is sitting on
      sprintf(buffer, "route add default dev %s", glbIfName);
      bcmSystem(buffer);
   }

   return pid;
}



/***************************************************************************
// Function Name: killAllApps().
// Description  : kill all available applications.
// Parameters   : none.
// Returns      : none.
****************************************************************************/
void killAllApps(int socketPid) {
   // NOTE: Add the app name before NULL. Order is important
   char *apps[]=
   {
      "bftpd",
      "telnetd",   
      "sshd",
      "snmp",
      "upnp",
      "reaim",
      "klogd",
      "syslogd",
      "tftpd",
      "ripd",
      "zebra ",
      "pppd",
      "dnsprobe",
      "dhcpc",
      "igmp",      
      NULL,
   };
   char cmd[CMD_LEN], app[CMD_LEN];
   char buffer[CMD_LEN];
   char pidStr[CMD_LEN];
   int pid = 0;
   int i = 0;
   FILE *fs;
   int curPid = getpid();

   bcmSystem("ps > /var/pslist");
   fs = fopen("/var/pslist", "r");
   bcmSystem("cat /var/pslist");

   // get ps line and parse for the pid for killing
   i = 0;
   do 
   {
      strcpy(app, apps[i]);
      while (fgets(buffer, CMD_LEN, fs) > 0) 
      {
         if (strstr(buffer, app) != NULL) // found command line with match app name
         {  
            // find pid string
            sscanf(buffer, "%s\n", pidStr);
            pid = atoi(pidStr);
            if (pid != curPid && pid != socketPid)
            {
               printf("kill %s process...\n", app);
               sprintf(cmd, "kill -9 %d", pid);
               bcmSystem(cmd);
            }
         }
      }
      rewind(fs);
   } while (apps[++i] != NULL);

   fclose(fs);
   bcmSystem("rm /var/pslist");
}

/***************************************************************************
// Function Name: getCrc32
// Description  : caculate the CRC 32 of the given data.
// Parameters   : pdata - array of data.
//                size - number of input data bytes.
//                crc - either CRC32_INIT_VALUE or previous return value.
// Returns      : crc.
****************************************************************************/
UINT32 getCrc32(byte *pdata, UINT32 size, UINT32 crc)
{
    while (size-- > 0)
        crc = (crc >> 8) ^ Crc32_table[(crc ^ *pdata++) & 0xff];

    return crc;
}


/***************************************************************************
// Function Name: verifyTag
// Description  : check if the image is brcm image or not. If passNumber == 1,
//                just check the tag crc for memory allocation.  If passNumber == 2, 
//                check the image crc and tagVersion.
// Parameters   : pTag and passNumber.
// Returns      : UPLOAD_OK or UPLOAD_FAIL_ILLEGAL_IMAGE
****************************************************************************/
UPLOAD_RESULT verifyTag(PFILE_TAG pTag, int passNumber)
{
    UINT32 crc;
    UPLOAD_RESULT status = UPLOAD_OK;
    int totalImageSize;
    int tagVer, curVer, tagVerLast;

    // check tag validate token first
    crc = CRC32_INIT_VALUE;
    crc = getCrc32((char*) pTag, (UINT32)TAG_LEN-TOKEN_LEN, crc);

    if (crc != (UINT32)(*(UINT32*)(pTag->tagValidationToken)))
    {
        printf("Not regular image file\n");
        return UPLOAD_FAIL_ILLEGAL_IMAGE;
    }
    if (passNumber == 1)
        return status;

    // if second pass, check imageCrc and tagVersion
    totalImageSize = atoi(pTag->totalImageLen);
    crc = CRC32_INIT_VALUE;
    crc = getCrc32(((char*)pTag + TAG_LEN), (UINT32) totalImageSize, crc);      
    if (crc != (UINT32) (*(UINT32*)(pTag->imageValidationToken)))
        return UPLOAD_FAIL_ILLEGAL_IMAGE;

    // only allow same or greater tag versions (up to tagVerLast - 26) to be uploaded
    tagVer = atoi(pTag->tagVersion);
    curVer = atoi(BCM_TAG_VER);
    tagVerLast = atoi(BCM_TAG_VER_LAST);

    if (tagVer < curVer || tagVer > tagVerLast)
    {
       printf("Firmware tag version [%d] is not compatible with the current Tag version [%d]\n", tagVer, curVer);
       return UPLOAD_FAIL_ILLEGAL_IMAGE;
    }

    return status;
}


// depending on the image type, do the brcm image or whole flash image
// return: UPLOAD_RESULT
UPLOAD_RESULT flashImage(char *imagePtr, PARSE_RESULT imageType, int imageLen)
{
    PFILE_TAG pTag = (PFILE_TAG) imagePtr;
    int cfeSize, rootfsSize, kernelSize;
    unsigned long cfeAddr, rootfsAddr, kernelAddr;
    UPLOAD_RESULT status = UPLOAD_OK;

    if (imageType == FLASH_IMAGE_FORMAT)  
    {
        printf("\nFlash whole image...");
        if ((sysFlashImageSet(imagePtr, imageLen-TOKEN_LEN, (int)FLASH_BASE, BCM_IMAGE_WHOLE)) == -1)
        {
            printf("Failed to flash whole image\n");
            return UPLOAD_FAIL_FLASH;
        }
        return UPLOAD_OK;
    }

    // check imageCrc
    cfeSize = rootfsSize = kernelSize = 0;

    // check cfe's existence
    cfeAddr = (unsigned long) strtoul(pTag->cfeAddress, NULL, 10);
    cfeSize = atoi(pTag->cfeLen);
    // check kernel existence
    kernelAddr = (unsigned long) strtoul(pTag->kernelAddress, NULL, 10);
    kernelSize = atoi(pTag->kernelLen);
    // check root filesystem existence
    rootfsAddr = (unsigned long) strtoul(pTag->rootfsAddress, NULL, 10);
    rootfsSize = atoi(pTag->rootfsLen);

    if (cfeAddr) 
    {
        printf("\nFlashing CFE...\n");
        if ((sysFlashImageSet(imagePtr+TAG_LEN, cfeSize, (int) cfeAddr,
            BCM_IMAGE_CFE)) == -1)
        {
            printf("Failed to flash CFE\n");
            return UPLOAD_FAIL_FLASH;
        }
    }

    if (rootfsAddr && kernelAddr) 
    {
        char *tagFs = imagePtr;

        // tag is alway at the sector start of fs
        if (cfeAddr)
        {
            tagFs = imagePtr + cfeSize;       // will trash cfe memory, but cfe is already flashed
            memcpy(tagFs, imagePtr, TAG_LEN);
        }

        printf("Flashing root file system and kernel...\n");
        if ((sysFlashImageSet(tagFs, TAG_LEN+rootfsSize+kernelSize, (int)(rootfsAddr-TAG_LEN), BCM_IMAGE_FS)) == -1)
        {
            printf("Failed to flash root file system\n");
            return UPLOAD_FAIL_FLASH;
        }
    }

    if (status != UPLOAD_OK)
        printf("**** IMAGE FLASH FAILED ****\n");
    else 
        printf("Image flash done.\n");
    
    return status;
}

// parseImageData
// parse the image data to see if it Broadcom flash format or flash image format file
// return: NO_IMAGE_FORMAT  
//         BROADCOM_IMAGE_FORMAT       
//         FLASH_IMAGE_FORMAT          
 
PARSE_RESULT parseImageData(char *image_start_ptr, int bufSize)
{
    PARSE_RESULT result = NO_IMAGE_FORMAT;

    // Check if the first part of the image is the Broadcom defined TAG record.
    if (verifyTag((FILE_TAG *) image_start_ptr, 2) == UPLOAD_OK)
    {
        result = BROADCOM_IMAGE_FORMAT;
        printf("Broadcom format verified.\n");
    }
    else
    {
        /* if It is not a Broadcom flash format file.  Now check if it is a
        * flash image format file.  A flash image format file must have a
        * CRC at the end of the image.
        */
        unsigned char *image_ptr = image_start_ptr;
        unsigned long image_len = bufSize - TOKEN_LEN;
        unsigned long crc = CRC32_INIT_VALUE;

        crc = getCrc32(image_ptr, image_len, crc);      
        if (memcmp(&crc, image_ptr + image_len, CRC_LEN) == 0)
        {
            printf("Flash image format [%ld bytes] verified.\n", image_len);
            result = FLASH_IMAGE_FORMAT;
        }
        else
            printf("Invalid image format [%d bytes].\n", bufSize);
    }

    return result;
}


/***************************************************************************
// Function Name: BcmScm_isServiceEnabled.
// Description  : check the given service is enabled.
// Parameters   : name -- the given service name.
//                where -- local or remote.
// Returns      : TRUE or FALSE.
****************************************************************************/
int BcmScm_isServiceEnabled(char *name, int where) {
   int enbl = FALSE, len = 0, lan = 0, wan = 0;
   char line[CLI_MAX_BUF_SZ];
   FILE* fs = fopen("/var/services.conf", "r");

   // if list is empty then return default which is
   // return true if access from lan and
   // return false if access from wan
   if ( fs == NULL ) {
      if ( where == CLI_ACCESS_LOCAL )
         return TRUE;
      else
         return FALSE;
   }

   len = strlen(name);

   while ( fscanf(fs, "%s\t%d\t%d", line, &lan, &wan) != EOF )
      if ( strncasecmp(line, name, len) == 0 ) {
         if ( where == CLI_ACCESS_LOCAL )
            enbl = lan;
         else if ( where == CLI_ACCESS_REMOTE )
            enbl = wan;
         break;
      }

   fclose(fs);

   return enbl;
}


/***************************************************************************
// Function Name: BcmScm_isInAccessControlList.
// Description  : check the given IP address in the access control list.
// Parameters   : addr -- the given IP address.
// Returns      : TRUE or FALSE.
****************************************************************************/
int BcmScm_isInAccessControlList(char * addr) {
   int in = TRUE, len = 0;
   char line[SEC_BUFF_MAX_LEN];
   FILE* fs = fopen("/var/acl.conf", "r");

   // if list is empty then return true
   // since allow to access for all by default
   if ( fs == NULL )
      return in;

   // get access control mode
   fgets(line, SEC_BUFF_MAX_LEN, fs);

   // if mode is disabled (zero) then return true
   // since allow to access for all by default
   if ( atoi(line) == 0 ) {
      fclose(fs);
      return in;
   }

   // mode is enabled then access control list is active
   // so in should be initialized to false
   in = FALSE;
   len = strlen(addr);

   while ( fgets(line, SEC_BUFF_MAX_LEN, fs) )
      if ( strncmp(line, addr, len) == 0 ) {
         in = TRUE;
         break;
      }

   fclose(fs);

   return in;
}

//**************************************************************************
// Function Name: bcmcheck_enable
// Description  : check the appName with ip address against the psi 
//                for access mode
// Parameters   : appName - application name in the acl.conf (telnet, ssh, etc.)
//                clntAddr - incoming ip address 
// Returns      : access mode - CLI_ACCESS_LOCAL, CLI_REMOTE_LOCAL, CLI_ACCESS_DISABLED
//**************************************************************************
int bcmCheckEnable(char *appName, struct in_addr clntAddr)
{
   // retrieve PSI from flash to make sure it's up-to-date
   //BcmPsi_profileRetrieve(0);

   // is client address in Access Control List ?
   if ( BcmScm_isInAccessControlList(inet_ntoa(clntAddr)) == FALSE )
      return CLI_ACCESS_DISABLED;

   if ( isAccessFromLan(clntAddr) == TRUE ) {
      // is enabled from lan ?
      if ( BcmScm_isServiceEnabled(appName, CLI_ACCESS_LOCAL) == FALSE )
         return CLI_ACCESS_DISABLED;
      else
         return CLI_ACCESS_LOCAL;
   } else {
      // is enabled from wan ?
      if ( BcmScm_isServiceEnabled(appName, CLI_ACCESS_REMOTE) == FALSE )
         return CLI_ACCESS_DISABLED;
      else
         return CLI_ACCESS_REMOTE;
   }
}
